Spark.Options.ValidationError exception (spark v2.6.0)

Copy Markdown View Source

An error that is returned (or raised) when options are invalid.

Since this is an exception, you can either raise it directly with raise/1 or turn it into a message string with Exception.message/1.

See %Spark.Options.ValidationError{} for documentation on the fields.

Summary

Types

t()

@type t() :: %Spark.Options.ValidationError{
  __exception__: true,
  key: atom(),
  keys_path: [atom()],
  message: term(),
  value: term()
}

Functions

%Spark.Options.ValidationError{}

(struct)

The error struct.

Only the following documented fields are considered public. All other fields are considered private and should not be referenced:

  • :key (atom/0) - The key that did not successfully validate.

  • :keys_path (list of atom/0) - If the key is nested, this is the path to the key.

  • :value (term/0) - The value that failed to validate. This field is nil if there was no value provided.